from functools import wraps
from typing import Callable
from typing import Optional

from pika import BasicProperties

from myrabbit.core.consumer.listener import Exchange
from myrabbit.core.consumer.listener import Listener
from myrabbit.core.consumer.listener import Queue
from myrabbit.core.consumer.pika_message import PikaMessage
from myrabbit.core.publisher.reconnecting_publisher import ReconnectingPublisherFactory
from myrabbit.core.serializer import JsonSerializer
from myrabbit.core.serializer import Serializer
from myrabbit.events.event_with_message import EventWithMessage
from myrabbit.events.listen_event_strategy.base import ListenEventStrategy
from myrabbit.events.listen_event_strategy.service_pool import ServicePool
from myrabbit.utils.functions import get_method_name


class EventBus:
    def __init__(
        self,
        amqp_url: str,
        serializer: Optional[Serializer] = None,
        default_exchange_params: Optional[dict] = None,
        default_queue_params: Optional[dict] = None,
    ):
        self._amqp_url = amqp_url
        self._serializer: Serializer = serializer or JsonSerializer()
        self.default_exchange_params = default_exchange_params or {}
        self.default_queue_params = default_queue_params or {}
        self._publisher_factory = ReconnectingPublisherFactory(self._amqp_url)

    def publish(
        self,
        event_source: str,
        event_name: str,
        body: Optional[dict] = None,
        properties: Optional[BasicProperties] = None,
    ) -> None:
        properties = properties or BasicProperties()

        body = body or {}
        content_type, binary_body = self._serializer.serialize(body)
        properties.content_type = content_type

        with self._publisher_factory.publisher() as publisher:
            publisher.publish(
                self._exchange(event_source),
                self._routing_key(event_name),
                binary_body,
                properties,
            )

    def listener(
        self,
        event_destination: str,
        event_source: str,
        event_name: str,
        callback: Callable[[EventWithMessage], None],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
        listen_strategy: Optional[ListenEventStrategy] = None,
        method_name: Optional[str] = None,
    ) -> Listener:
        listen_strategy = listen_strategy or ServicePool()

        method_name = get_method_name(method_name, callback)

        queue_params = queue_params or {}
        queue_params = {**self.default_queue_params, **queue_params}
        queue_params.setdefault(
            "name",
            listen_strategy.get_queue_name(
                event_destination, event_source, event_name, method_name,
            ),
        )

        exchange_params = exchange_params or {}
        exchange_params = {**self.default_exchange_params, **exchange_params}
        exchange_params.setdefault("type", "topic")
        exchange_params.setdefault("name", self._exchange(event_source))

        @wraps(callback)
        def deserialize_message(message: PikaMessage) -> None:
            callback(
                EventWithMessage(self._serializer.deserialize(message.body), message)
            )

        return Listener(
            exchange=Exchange(**exchange_params),
            queue=Queue(**queue_params),
            routing_key=event_name,
            handle_message=deserialize_message,
        )

    def _exchange(self, event_source: str) -> str:
        return f"{event_source}.events"

    def _routing_key(self, event: str) -> str:
        return event



from __future__ import annotations

import logging
from contextlib import contextmanager
from functools import wraps
from types import TracebackType
from typing import Any
from typing import Callable
from typing import Generator
from typing import Optional
from typing import Type

import pika
from pika import BasicProperties
from pika import URLParameters
from pika.adapters.blocking_connection import BlockingChannel
from pika.exceptions import ChannelClosedByBroker

from myrabbit.core.consumer.pika_message import PikaMessage
from myrabbit.core.publisher.rpc import rpc

logger = logging.getLogger(__name__)


def ignore_missing_exchange(fn: Callable) -> Callable:
    @wraps(fn)
    def handle(*args: Any, **kwargs: Any) -> Any:
        try:
            return fn(*args, **kwargs)
        except ChannelClosedByBroker as e:
            if int(e.reply_code) == 404:
                logger.error(
                    "Can not publish message because exchange does not exist: %s %s",
                    e.reply_code,
                    e.reply_text,
                )
            else:
                raise

    return handle


class Publisher:
    def __init__(self, connection: pika.BlockingConnection):
        self._connection = connection
        self._channel: BlockingChannel = self._connection.channel()

    @ignore_missing_exchange
    def publish(
        self,
        exchange: str,
        routing_key: str,
        message: bytes,
        properties: Optional[BasicProperties] = None,
    ) -> None:
        self._channel.basic_publish(
            exchange, routing_key, message, properties,
        )

    @ignore_missing_exchange
    def close(self) -> None:
        self._channel.close()
        self._connection.close()

    def rpc(
        self,
        exchange: str,
        routing_key: str,
        message: bytes,
        properties: Optional[BasicProperties] = None,
        timeout: Optional[int] = 1,
    ) -> PikaMessage:
        return rpc(
            self._connection, exchange, routing_key, message, properties, timeout
        )

    def __enter__(self) -> Publisher:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[Exception]],
        exc_val: Optional[Any],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.close()


@contextmanager
def make_publisher(amqp_url: str) -> Generator[Publisher, None, None]:
    with pika.BlockingConnection(URLParameters(amqp_url)) as conn:
        with Publisher(conn) as p:
            yield p



import abc
from typing import Any
from typing import Tuple


ContentType = str


class Serializer(abc.ABC):
    @abc.abstractmethod
    def serialize(self, data: Any) -> Tuple[ContentType, bytes]:
        pass

    @abc.abstractmethod
    def deserialize(self, data: bytes) -> Any:
        pass



from .base import ListenEventStrategy


class ServicePool(ListenEventStrategy):
    def get_queue_name(
        self,
        event_destination: str,
        event_source: str,
        event_name: str,
        method_name: str,
    ) -> str:
        return f"{event_source}.{event_name}.to.{event_destination}.{method_name}"



import logging
import uuid
from functools import wraps
from typing import Any
from typing import Callable
from typing import Optional
from typing import Union

import pika
from pika import BasicProperties
from pika import URLParameters

from myrabbit.commands import command_outcome
from myrabbit.commands.command_outcome import CommandReply
from myrabbit.commands.command_with_message import CommandWithMessage
from myrabbit.commands.command_with_message import ReplyWithMessage
from myrabbit.core.consumer.listener import Exchange
from myrabbit.core.consumer.listener import Listener
from myrabbit.core.consumer.listener import Queue
from myrabbit.core.consumer.pika_message import PikaMessage
from myrabbit.core.consumer.reply import Reply
from myrabbit.core.publisher.publisher import Publisher
from myrabbit.core.publisher.reconnecting_publisher import ReconnectingPublisherFactory
from myrabbit.core.serializer import JsonSerializer
from myrabbit.core.serializer import Serializer

logger = logging.getLogger(__name__)


class CommandBus:
    def __init__(
        self,
        amqp_url: str,
        serializer: Optional[Serializer] = None,
        default_exchange_params: Optional[dict] = None,
        default_queue_params: Optional[dict] = None,
    ):
        self._amqp_url = amqp_url
        self._serializer: Serializer = serializer or JsonSerializer()
        self.default_exchange_params = default_exchange_params or {}
        self.default_queue_params = default_queue_params or {}
        self._publisher_factory = ReconnectingPublisherFactory(self._amqp_url)

    def send(
        self,
        command_sender: str,
        command_destination: str,
        command_name: str,
        body: Optional[dict] = None,
        properties: Optional[BasicProperties] = None,
    ) -> None:
        properties = properties or BasicProperties()

        body = body or {}
        content_type, binary_body = self._serializer.serialize(body)

        properties.content_type = content_type

        if properties.reply_to is None:
            properties.reply_to = self._default_reply_queue_name(
                command_sender, command_destination, command_name
            )

        if properties.correlation_id is None:
            properties.correlation_id = uuid.uuid4().hex

        with self._publisher_factory.publisher() as publisher:
            publisher.publish(
                self._exchange(command_destination),
                self._routing_key(command_name),
                binary_body,
                properties,
            )

    def listener(
        self,
        command_destination: str,
        command_name: str,
        callback: Callable[[CommandWithMessage], Optional[Union[CommandReply, Any]]],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Listener:
        queue_params = queue_params or {}
        queue_params = {**self.default_queue_params, **queue_params}
        queue_params.setdefault(
            "name", f"{command_destination}.{command_name}",
        )

        exchange_params = exchange_params or {}
        exchange_params = {**self.default_exchange_params, **exchange_params}
        exchange_params.setdefault("type", "direct")
        exchange_params.setdefault("name", self._exchange(command_destination))

        @wraps(callback)
        def deserialize_command_and_handle_reply(
            message: PikaMessage,
        ) -> Optional[Reply]:
            try:
                callback_result: Optional[Union[CommandReply, Any]] = callback(
                    CommandWithMessage(
                        self._serializer.deserialize(message.body), message
                    )
                )
            except Exception as e:
                logger.exception(
                    "Exception happened during handling message %s using handler %s",
                    message,
                    callback,
                )
                return command_outcome.exception(e).to_reply(self._serializer)

            if callback_result is None:
                return None

            if isinstance(callback_result, CommandReply):
                return callback_result.to_reply(self._serializer)

            return command_outcome.success(body=callback_result).to_reply(
                self._serializer
            )

        return Listener(
            exchange=Exchange(**exchange_params),
            queue=Queue(**queue_params),
            routing_key=self._routing_key(command_name),
            handle_message=deserialize_command_and_handle_reply,
        )

    def reply_listener(
        self,
        command_sender: str,
        command_destination: str,
        command_name: str,
        callback: Callable[[ReplyWithMessage], None],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Listener:
        queue_params = queue_params or {}
        queue_params = {**self.default_queue_params, **queue_params}
        queue_params.setdefault(
            "name",
            self._default_reply_queue_name(
                command_sender, command_destination, command_name
            ),
        )
        routing_key = queue_params["name"]  # Exchange is direct.

        exchange_params = exchange_params or {}
        exchange_params = {**self.default_exchange_params, **exchange_params}
        exchange_params.setdefault("type", "direct")
        exchange_params.setdefault("name", self._exchange(command_destination))

        @wraps(callback)
        def deserialize_message(message: PikaMessage) -> None:
            reply: dict = self._serializer.deserialize(message.body)
            callback(ReplyWithMessage(reply=reply, message=message))

        return Listener(
            exchange=Exchange(**exchange_params),
            queue=Queue(**queue_params),
            routing_key=routing_key,
            handle_message=deserialize_message,
        )

    def _exchange(self, command_destination: str) -> str:
        return f"{command_destination}.commands"

    def _routing_key(self, command_name: str) -> str:
        return command_name

    def _default_reply_queue_name(
        self, command_sender: str, command_destination: str, command_name: str
    ) -> str:
        return (
            f"{command_sender}.listen-reply-from:{command_destination}.{command_name}"
        )



import functools
import logging
from concurrent.futures.thread import ThreadPoolExecutor
from functools import partial
from typing import Dict
from typing import List
from typing import Optional

import pika
from pika import SelectConnection
from pika.adapters.asyncio_connection import AsyncioConnection
from pika.channel import Channel
from pika.connection import Connection
from pika.spec import Basic
from pika.spec import Exchange
from pika.spec import Queue

from myrabbit.core.consumer.channel import ConsumedChannel
from myrabbit.core.consumer.listener import Listener
from myrabbit.core.consumer.pika_message import PikaMessage

logger = logging.getLogger(__name__)


class Consumer(object):
    """This is an example consumer that will handle unexpected interactions
    with RabbitMQ such as channel and connection closures.
    If RabbitMQ closes the connection, this class will stop and indicate
    that reconnection is necessary. You should look at the output, as
    there are limited reasons why the connection may be closed, which
    usually are tied to permission related issues or socket timeouts.
    If the channel is closed, it will indicate a problem with one of the
    commands that were issued and that should surface in the output as well.
    """

    def __init__(
        self, amqp_url: str, listeners: List[Listener], prefetch_count: int = 1
    ):
        """Create a new instance of the consumer class, passing in the AMQP
        URL used to connect to RabbitMQ.
        :param str amqp_url: The AMQP url to connect with
        """
        self.should_reconnect = False
        self.was_consuming = False

        self._listeners: List[Listener] = listeners
        self._channels: Dict[int, ConsumedChannel] = {}

        self._connection = None
        self._closing = False
        self._url = amqp_url
        self._consuming = False
        # In production, experiment with higher prefetch values
        # for higher consumer throughput
        self._prefetch_count = prefetch_count

    def connect(self) -> SelectConnection:
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.
        :rtype: pika.adapters.asyncio_connection.AsyncioConnection
        """
        logger.info("Connecting to %s", self._url)
        return SelectConnection(
            parameters=pika.URLParameters(self._url),
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed,
        )

    def maybe_close_connection(self) -> None:
        if not self._channels:
            self.close_connection()

    def close_connection(self) -> None:
        assert self._connection

        self._consuming = False
        if self._connection.is_closing or self._connection.is_closed:
            logger.info("Connection is closing or already closed")
        else:
            logger.info("Closing connection")
            self._connection.close()

    def on_connection_open(self, _unused_connection: SelectConnection) -> None:
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.
        """
        logger.info("Connection opened")

        for listener in self._listeners:
            self.open_channel(listener)

    def on_connection_open_error(
        self, _unused_connection: AsyncioConnection, err: Exception
    ) -> None:
        """This method is called by pika if the connection to RabbitMQ
        can't be established.
        """
        logger.error("Connection open failed: %s", err)
        self.reconnect()

    def on_connection_closed(
        self, _unused_connection: Connection, reason: Exception
    ) -> None:
        """
        This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.
        """
        assert self._connection

        if self._closing:
            self._connection.ioloop.stop()
        else:
            logger.warning("Connection closed, reconnect necessary: %s", reason)
            self.reconnect()

    def reconnect(self) -> None:
        """
        Will be invoked if the connection can't be opened or is
        closed. Indicates that a reconnect is necessary then stops the
        ioloop.
        """
        self.should_reconnect = True
        self.stop()

    def open_channel(self, listener: Listener) -> None:
        """Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.
        """
        assert self._connection
        logger.info("Creating a new channel")
        self._connection.channel(
            on_open_callback=partial(self.on_channel_open, listener=listener)
        )

    def on_channel_open(self, channel: Channel, listener: Listener) -> None:
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.
        Since the channel is now open, we'll declare the exchange to use.
        :param pika.channel.Channel channel: The channel object
        """
        logger.info("Channel opened")
        consumed_channel = ConsumedChannel(listener=listener, pika_channel=channel)
        self.remember_channel(consumed_channel)

        self.add_on_channel_close_callback(consumed_channel)
        self.setup_exchange(consumed_channel)

    def remember_channel(self, channel: ConsumedChannel) -> None:
        self._channels[int(channel.pika_channel)] = channel

    def forget_channel(self, channel: ConsumedChannel) -> None:
        self._channels.pop(int(channel.pika_channel), None)

    def add_on_channel_close_callback(self, consumed_channel: ConsumedChannel) -> None:
        """This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.
        """
        consumed_channel.pika_channel.add_on_close_callback(
            partial(self.on_channel_closed, consumed_channel=consumed_channel)
        )

    def on_channel_closed(
        self, channel: Channel, reason: Exception, consumed_channel: ConsumedChannel
    ) -> None:
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.
        """
        logger.warning("Channel %i was closed: %s", channel, reason)
        self.forget_channel(consumed_channel)
        self.maybe_close_connection()

    def setup_exchange(self, consumed_channel: ConsumedChannel) -> None:
        """
        Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
        command. When it is complete, the on_exchange_declareok method will
        be invoked by pika.
        """
        logger.info("Declaring exchange: %s", consumed_channel.exchange)
        exchange = consumed_channel.exchange

        if exchange.name == "":
            self.setup_queue(consumed_channel)
            return

        consumed_channel.pika_channel.exchange_declare(
            exchange=exchange.name,
            exchange_type=exchange.type,
            callback=functools.partial(
                self.on_exchange_declareok, channel=consumed_channel
            ),
            durable=exchange.durable,
            auto_delete=exchange.auto_delete,
        )

    def on_exchange_declareok(
        self, _unused_frame: Exchange.DeclareOk, channel: ConsumedChannel
    ) -> None:
        """
        Invoked by pika when RabbitMQ has finished the
        Exchange.Declare RPC command.
        """
        logger.info("Exchange declared: %s", channel.exchange)
        self.setup_queue(channel)

    def setup_queue(self, channel: ConsumedChannel) -> None:
        """
        Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
        command. When it is complete, the on_queue_declareok method will
        be invoked by pika.
        """
        logger.info("Declaring queue %s", channel.queue)
        cb = functools.partial(self.on_queue_declareok, channel=channel)
        queue = channel.queue
        channel.pika_channel.queue_declare(
            queue=queue.name,
            durable=queue.durable,
            exclusive=queue.exclusive,
            auto_delete=queue.auto_delete,
            callback=cb,
        )

    def on_queue_declareok(
        self, _unused_frame: Queue.DeclareOk, channel: ConsumedChannel
    ) -> None:
        """
        Method invoked by pika when the Queue.Declare RPC call made in
        setup_queue has completed. In this method we will bind the queue
        and exchange together with the routing key by issuing the Queue.Bind
        RPC command. When this command is complete, the on_bindok method will
        be invoked by pika.
        """
        if channel.exchange.name == "":
            self.set_qos(channel)
            return

        logger.info(
            "Binding %s to %s with routing key %s",
            channel.exchange,
            channel.queue,
            channel.listener.routing_key,
        )
        cb = functools.partial(self.on_bindok, channel=channel)
        channel.pika_channel.queue_bind(
            channel.queue.name,
            channel.exchange.name,
            routing_key=channel.listener.routing_key,
            callback=cb,
        )

    def on_bindok(self, _unused_frame: Queue.BindOk, channel: ConsumedChannel) -> None:
        """
        Invoked by pika when the Queue.Bind method has completed. At this
        point we will set the prefetch count for the channel.
        """
        logger.info("Queue bound: %s", channel.queue)
        self.set_qos(channel)

    def set_qos(self, channel: ConsumedChannel) -> None:
        """
        This method sets up the consumer prefetch to only be delivered
        one message at a time. The consumer must acknowledge this message
        before RabbitMQ will deliver another one. You should experiment
        with different prefetch values to achieve desired performance.
        """
        channel.pika_channel.basic_qos(
            prefetch_count=self._prefetch_count,
            callback=partial(self.on_basic_qos_ok, channel=channel),
        )

    def on_basic_qos_ok(
        self, _unused_frame: Basic.QosOk, channel: ConsumedChannel
    ) -> None:
        """
        Invoked by pika when the Basic.QoS method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.
        :param pika.frame.Method _unused_frame: The Basic.QosOk response frame
        """
        logger.info("QOS set to: %d", self._prefetch_count)
        self.start_consuming(channel)

    def start_consuming(self, channel: ConsumedChannel) -> None:
        """
        This method sets up the consumer by first calling
        add_on_cancel_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming. The on_message method is passed in as a callback pika
        will invoke when a message is fully received.
        """
        logger.info("Issuing consumer related RPC commands")
        self.add_on_cancel_callback(channel)

        consumer_tag = channel.pika_channel.basic_consume(
            queue=channel.queue.name,
            on_message_callback=partial(self.on_message, channel=channel),
            auto_ack=channel.listener.auto_ack,
        )
        channel.consumer_tag = consumer_tag

        self.was_consuming = True
        self._consuming = True

    def add_on_cancel_callback(self, channel: ConsumedChannel) -> None:
        """Add a callback that will be invoked if RabbitMQ cancels the consumer
        for some reason. If RabbitMQ does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.
        """
        channel.pika_channel.add_on_cancel_callback(
            partial(self.on_consumer_cancelled, channel=channel)
        )

    def on_consumer_cancelled(
        self, method_frame: Basic.Cancel, channel: ConsumedChannel
    ) -> None:
        """
        Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.
        """
        logger.info("Consumer was cancelled remotely, shutting down: %r", method_frame)
        if channel.pika_channel:
            channel.pika_channel.close()
            self.forget_channel(channel)

    def on_message(
        self,
        unused_channel: Channel,
        basic_deliver: Basic.Deliver,
        properties: pika.BasicProperties,
        body: bytes,
        channel: ConsumedChannel,
    ) -> None:
        """
        Invoked by pika when a message is delivered from RabbitMQ. The
        channel is passed for your convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.
        """
        logger.info(
            "Received message #%s from %s: %s (consumer %s, corr_id: %s)",
            basic_deliver.delivery_tag,
            properties.app_id,
            body,
            channel.consumer_tag,
            properties.correlation_id,
        )
        self._handle_message(unused_channel, basic_deliver, properties, body, channel)

    def _handle_message(
        self,
        unused_channel: Channel,
        basic_deliver: Basic.Deliver,
        properties: pika.BasicProperties,
        body: bytes,
        channel: ConsumedChannel,
    ) -> None:
        channel.listener.handle(
            PikaMessage(unused_channel, basic_deliver, properties, body),
        )

    def stop_consuming(self) -> None:
        """Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.
        """
        for channel in self._channels.values():
            logger.info("Sending a Basic.Cancel RPC command to RabbitMQ")
            cb = functools.partial(self.on_cancelok, channel=channel)
            channel.pika_channel.basic_cancel(channel.consumer_tag, cb)

    def on_cancelok(
        self, _unused_frame: Basic.CancelOk, channel: ConsumedChannel
    ) -> None:
        """
        This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.
        """
        if not self._channels:
            self._consuming = False

        logger.info(
            "RabbitMQ acknowledged the cancellation of the consumer: %s",
            channel.consumer_tag,
        )
        self.close_channel(channel)

    def close_channel(self, channel: ConsumedChannel) -> None:
        """
        Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.
        """
        logger.info("Closing the channel")
        channel.pika_channel.close()

    def run(self) -> None:
        """Run the example consumer by connecting to RabbitMQ and then
        starting the IOLoop to block and allow the AsyncioConnection to operate.
        """
        self._connection = self.connect()
        assert self._connection
        self._connection.ioloop.start()

    def stop(self) -> None:
        """Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.
        """
        assert self._connection
        if not self._closing:
            self._closing = True
            logger.info("Stopping")
            if self._consuming:
                self.stop_consuming()
                try:
                    self._connection.ioloop.start()
                except RuntimeError:
                    pass
            else:
                self._connection.ioloop.stop()
            logger.info("Stopped")


class ThreadedConsumer(Consumer):
    def __init__(  # type: ignore
        self, *args, executor: Optional[ThreadPoolExecutor] = None, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self._executor = executor or ThreadPoolExecutor()

    def _handle_message(
        self,
        unused_channel: Channel,
        basic_deliver: Basic.Deliver,
        properties: pika.BasicProperties,
        body: bytes,
        channel: ConsumedChannel,
    ) -> None:
        self._executor.submit(
            channel.listener.handle,
            PikaMessage(unused_channel, basic_deliver, properties, body),
        )

    def stop(self) -> None:
        super().stop()
        self._executor.shutdown()



from dataclasses import asdict
from dataclasses import is_dataclass
from typing import Tuple
from typing import Type

from .converter import Converter
from .converter import T


class DataclassConverter(Converter):
    def name_and_body(self, model: T) -> Tuple[str, dict]:
        return model.__class__.__name__, asdict(model)

    def name(self, model: Type[T]) -> str:
        return model.__name__

    def instantiate(self, model_class: Type[T], body: dict) -> T:
        return model_class(**body)  # type: ignore

    def accepts(self, model: T) -> bool:
        return is_dataclass(model)



from dataclasses import dataclass
from typing import Generic
from typing import TypeVar

from myrabbit.commands.command_outcome import CommandOutcome
from myrabbit.commands.reply_headers import CommandReplyHeaders
from myrabbit.core.consumer.pika_message import PikaMessage

CommandType = TypeVar("CommandType")
CommandReplyType = TypeVar("CommandReplyType")


@dataclass
class CommandWithMessage(Generic[CommandType]):
    command: CommandType
    message: PikaMessage


@dataclass
class ReplyWithMessage(Generic[CommandReplyType]):
    reply: CommandReplyType
    message: PikaMessage

    def is_success(self) -> bool:
        reply_outcome: str = self.message.properties.headers[
            CommandReplyHeaders.REPLY_OUTCOME
        ]
        return reply_outcome == CommandOutcome.SUCCESS.name

    def is_failure(self) -> bool:
        reply_outcome: str = self.message.properties.headers[
            CommandReplyHeaders.REPLY_OUTCOME
        ]
        return reply_outcome == CommandOutcome.FAILURE.name



import traceback
from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import Optional

import pika

from myrabbit.core.consumer.reply import Reply
from myrabbit.commands.reply_headers import CommandReplyHeaders
from myrabbit.core.serializer import Serializer


@dataclass
class CommandReply:
    body: Any
    properties: Optional[pika.BasicProperties] = None

    def to_reply(self, serializer: Serializer) -> Reply:
        content_type, body = serializer.serialize(self.body)
        properties = self.properties
        return Reply(body=body, properties=properties)


@dataclass
class Success:
    pass


@dataclass
class Failure:
    pass


class CommandOutcome(Enum):
    SUCCESS = Success()
    FAILURE = Failure()


@dataclass
class ExceptionInfo:
    type: str
    message: str
    traceback: str


def make_command_reply(
    body: Any, outcome: CommandOutcome, properties: pika.BasicProperties
) -> CommandReply:
    properties = properties or pika.BasicProperties()

    headers = {
        CommandReplyHeaders.REPLY_OUTCOME: outcome.name,
        CommandReplyHeaders.REPLY_TYPE: body.__class__.__name__,
    }
    properties.headers = properties.headers or {}
    properties.headers.update(headers)
    return CommandReply(body=body, properties=properties)


def success(
    body: Any = CommandOutcome.SUCCESS.value,
    properties: Optional[pika.BasicProperties] = None,
) -> CommandReply:
    return make_command_reply(body, CommandOutcome.SUCCESS, properties)


def failure(
    body: Any = CommandOutcome.FAILURE.value,
    properties: Optional[pika.BasicProperties] = None,
) -> CommandReply:
    return make_command_reply(body, CommandOutcome.FAILURE, properties)


def exception(exc: Exception) -> CommandReply:
    return failure(
        ExceptionInfo(
            type=type(exc).__name__, message=str(exc), traceback=traceback.format_exc(),
        )
    )



import logging
import threading
from typing import Optional

import pika
from pika import BasicProperties
from pika.adapters.blocking_connection import BlockingChannel
from pika.channel import Channel
from pika.exceptions import ChannelWrongStateError
from pika.exceptions import ConnectionWrongStateError
from pika.spec import Basic

from myrabbit.core.consumer.pika_message import PikaMessage

LOGGER = logging.getLogger(__name__)


def _timeout_timer(
    chan: Channel,
    wait_secs: int,
    stop_timer_event: threading.Event,
    timeout_event: threading.Event,
) -> None:
    # Lets current thread to block while waiting for timeout.
    blocker = threading.Event()

    # Wakeup ten times to check if timer is still needed to run.
    sleep_time = wait_secs / 10
    for _ in range(10):
        blocker.wait(timeout=sleep_time)
        if stop_timer_event.is_set():
            return

    def run_threadsafe() -> None:
        try:
            # https://www.rabbitmq.com/amqp-0-9-1-reference.html#constants
            internal_error = 541
            chan.close(
                reply_code=internal_error,
                reply_text=f"RPC wait timeout reached ({wait_secs} s)",
            )
        except (ChannelWrongStateError, ConnectionWrongStateError):
            LOGGER.debug("Rpc call aborted because of timeout (%s s)", wait_secs)
        except Exception:
            LOGGER.exception("Exception happened during rpc call timeout handler")
        else:
            LOGGER.debug("Rpc call aborted due to timeout (%s s)", wait_secs)

        timeout_event.set()

    try:
        chan.connection.add_callback_threadsafe(run_threadsafe)
    except ConnectionWrongStateError:
        pass


def rpc(
    connection: pika.BlockingConnection,
    exchange_name: str,
    routing_key: str,
    body: bytes,
    properties: Optional[BasicProperties] = None,
    timeout: Optional[int] = None,
) -> PikaMessage:
    response = None
    timeout_event = threading.Event()
    stop_timer_event = threading.Event()

    def on_reply_received(
        reply_channel: Channel,
        reply_method_frame: Basic.Deliver,
        reply_properties: pika.BasicProperties,
        reply_body: bytes,
    ) -> None:
        nonlocal response
        response = PikaMessage(
            reply_channel, reply_method_frame, reply_properties, reply_body
        )
        reply_channel.close()
        stop_timer_event.set()

    direct_reply_queue = "amq.rabbitmq.reply-to"
    channel: BlockingChannel

    with connection.channel() as channel:
        channel.basic_consume(direct_reply_queue, on_reply_received, auto_ack=True)

        props = properties or BasicProperties()
        props.reply_to = direct_reply_queue

        channel.basic_publish(
            exchange=exchange_name,
            routing_key=routing_key,
            body=body,
            properties=props,
        )

        if timeout:
            timer = threading.Thread(
                target=_timeout_timer,
                args=(channel, timeout, stop_timer_event, timeout_event),
                daemon=True,
            )
            timer.start()

        channel.start_consuming()

    if timeout_event.is_set():
        raise TimeoutError

    assert isinstance(response, PikaMessage)
    return response



import inspect
from typing import Callable
from typing import Optional


def get_method_name(method_name: Optional[str], fn: Callable) -> str:
    if inspect.ismethod(fn.__call__):
        return method_name or fn.__class__.__name__
    return method_name or fn.__name__



import uuid
from typing import Optional

from .base import ListenEventStrategy


class Broadcast(ListenEventStrategy):
    def __init__(self, broadcast_identifier: Optional[str] = None):
        self._broadcast_identifier = broadcast_identifier or uuid.uuid4().hex

    def get_queue_name(
        self,
        event_destination: str,
        event_source: str,
        event_name: str,
        method_name: str,
    ) -> str:
        return (
            f"{event_source}.{event_name}.to."
            f"{event_destination}.{method_name}:{self._broadcast_identifier}"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['myrabbit',
 'myrabbit.commands',
 'myrabbit.core',
 'myrabbit.core.consumer',
 'myrabbit.core.converter',
 'myrabbit.core.publisher',
 'myrabbit.core.serializer',
 'myrabbit.events',
 'myrabbit.events.listen_event_strategy',
 'myrabbit.utils']

package_data = \
{'': ['*']}

install_requires = \
['orjson>=3.0.0,<4.0.0', 'pika>=1.1,<2.0', 'pydantic>=1.4,<2.0']

setup_kwargs = {
    'name': 'myrabbit',
    'version': '0.1.3',
    'description': 'Python library for microservice messaging via RabbitMq',
    'long_description': None,
    'author': 'asyncee',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
import time
from typing import Type

from myrabbit.core.consumer.consumer import Consumer

logger = logging.getLogger(__name__)


class ReconnectingConsumer:
    """
    Consumer that automatically reconnects with increasing delay.
    """

    def __init__(self, consumer_cls: Type[Consumer], consumer_kwargs: dict) -> None:
        self._reconnect_delay = 0
        self._consumer_cls = consumer_cls
        self._consumer_kwargs = consumer_kwargs

        self._consumer = self.make_consumer()
        self._should_run = True

    def make_consumer(self) -> Consumer:
        return self._consumer_cls(**self._consumer_kwargs)

    def run(self) -> None:
        while self._should_run:
            try:
                self._consumer.run()
            except KeyboardInterrupt:
                self._consumer.stop()
                break
            self._maybe_reconnect()

    def stop(self) -> None:
        self._consumer.stop()
        self._should_run = False

    def _maybe_reconnect(self) -> None:
        if self._consumer.should_reconnect:
            self._consumer.stop()
            reconnect_delay = self._get_reconnect_delay()
            logger.info("Reconnecting after %d seconds", reconnect_delay)
            time.sleep(reconnect_delay)
            self._consumer = self.make_consumer()

    def _get_reconnect_delay(self) -> int:
        if self._consumer.was_consuming:
            self._reconnect_delay = 0
        else:
            self._reconnect_delay += 1
        if self._reconnect_delay > 30:
            self._reconnect_delay = 30
        return self._reconnect_delay



from typing import Callable
from typing import List
from typing import Optional
from typing import Type

from pika import BasicProperties

from myrabbit import EventBus
from myrabbit import EventBusAdapter
from myrabbit.commands.command_bus import CommandBus
from myrabbit.commands.command_bus_adapter import CommandBusAdapter
from myrabbit.commands.command_with_message import CommandReplyType
from myrabbit.commands.command_with_message import CommandType
from myrabbit.commands.command_with_message import ReplyWithMessage
from myrabbit.core.consumer.listener import Listener
from myrabbit.events.event_with_message import EventType
from myrabbit.events.listen_event_strategy import ListenEventStrategy


class Service:
    def __init__(self, service_name: str, event_bus: EventBus, command_bus: CommandBus):
        self.service_name = service_name
        self._event_bus = event_bus
        self._event_bus_adapter = self._make_event_bus_adapter(event_bus)
        self._command_bus = command_bus
        self._command_bus_adapter = self._make_command_bus_adapter(command_bus)

        self._listeners: List[Listener] = []

    @property
    def event_bus(self) -> EventBus:
        return self._event_bus

    @property
    def command_bus(self) -> CommandBus:
        return self._command_bus

    @property
    def listeners(self) -> List[Listener]:
        return self._listeners

    def publish(
        self, event: EventType, properties: Optional[BasicProperties] = None
    ) -> None:
        properties = properties or BasicProperties()
        properties.app_id = self.service_name
        self._event_bus_adapter.publish(
            event_source=self.service_name, event=event, properties=properties
        )

    def send(
        self,
        command_destination: str,
        command: CommandType,
        properties: Optional[BasicProperties] = None,
        reply_to: Optional[str] = None,
    ) -> None:
        properties = properties or BasicProperties()
        properties.app_id = self.service_name

        if reply_to is not None:
            properties.reply_to = reply_to

        self._command_bus_adapter.send(
            self.service_name,
            command_destination=command_destination,
            command=command,
            properties=properties,
        )

    def on_event(
        self,
        event_source: str,
        event_type: Type[EventType],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
        listen_strategy: Optional[ListenEventStrategy] = None,
        method_name: Optional[str] = None,
    ) -> Callable:
        def register_event_listener(fn: Callable) -> Callable:
            self._listeners.append(
                self._event_bus_adapter.listener(
                    event_destination=self.service_name,
                    event_source=event_source,
                    event_type=event_type,
                    callback=fn,
                    exchange_params=exchange_params,
                    queue_params=queue_params,
                    listen_strategy=listen_strategy,
                    method_name=method_name,
                )
            )
            return fn

        return register_event_listener

    def on_command(
        self,
        command_type: Type[CommandType],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Callable:
        def register_command_listener(fn: Callable) -> Callable:
            self._listeners.append(
                self._command_bus_adapter.listener(
                    command_destination=self.service_name,
                    command_type=command_type,
                    callback=fn,
                    exchange_params=exchange_params,
                    queue_params=queue_params,
                )
            )
            return fn

        return register_command_listener

    def on_command_reply(
        self,
        command_destination: str,
        command_type: Type[CommandType],
        reply_type: Optional[Type[CommandReplyType]] = None,
        listen_on: Optional[str] = None,
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Callable:
        if listen_on:
            queue_params = queue_params or {}
            queue_params["name"] = listen_on

        def register_command_listener(
            fn: Callable[[ReplyWithMessage], None]
        ) -> Callable:
            self._listeners.append(
                self._command_bus_adapter.reply_listener(
                    command_sender=self.service_name,
                    command_destination=command_destination,
                    command_type=command_type,
                    callback=fn,
                    reply_type=reply_type,
                    exchange_params=exchange_params,
                    queue_params=queue_params,
                )
            )
            return fn

        return register_command_listener

    def _make_event_bus_adapter(self, event_bus: EventBus) -> EventBusAdapter:
        return EventBusAdapter(event_bus)

    def _make_command_bus_adapter(self, command_bus: CommandBus) -> CommandBusAdapter:
        return CommandBusAdapter(command_bus)
